# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 16:07:59 2021
@author: Anna Rauch

This script contains a set of general functions/methods usually related to
linear algebra and connectivty issues. These functions are independent from
the TIE-method, yet are used by it.
"""
import math
import numpy as np
import platform
import tkinter as tk
from tkinter import messagebox


def angle2normal(azim, dip):
    """
    Calculate the normal vector of a plane (or set of planes) defined with orientational angles in degrees.

    Parameters
    ----------
    azim : float or array_like
        Angles of azimuth (dip azimuth) in degrees.
    dip : float or array_like
        Dip angles of the plane(s) in degrees.

    Returns
    -------
    numpy.ndarray
        Normal vector of the plane(s) in the form [normalx, normaly, normalz].
    """

    dip = np.array(dip) * np.pi / 180
    azim = np.array(azim) * np.pi / 180

    if np.size(dip) == 1:
        dip = [dip]
        azim = [azim]

    normalz = [np.cos(dp) for dp in dip]
    normalx = [np.sin(dip[ind]) * np.sin(azim[ind]) for ind in range(np.size(dip))]
    normaly = [np.sin(dip[ind]) * np.cos(azim[ind]) for ind in range(np.size(dip))]

    return np.array([normalx, normaly, normalz]).T


def angle2vect(trend, plunge):
    """
    Calculate the directional vector (length = 1) from a line defined by angles - trend and plunge.

    Parameters
    ----------
    trend : float or array_like
        Angles of trend (plunge azimuth) in degrees.
    plunge : float or array_like
        Plunge angles of the directional vector in degrees.

    Returns
    -------
    tuple of numpy.ndarray
        Coordinates of the oriented vector in the form (vx, vy, vz).
    """

    # Convert angles to radians
    trend = np.array(trend) * np.pi / 180
    plunge = np.array(plunge) * np.pi / 180

    # Handle single values or arrays
    if np.size(trend) == 1:
        trend = [trend]
        plunge = [plunge]

    # Calculate coordinates of the oriented vector
    vz = [-np.sin(pl) for pl in plunge]
    vx = [np.sin(trend[k]) * np.cos(plunge[k]) for k in range(np.size(trend))]
    vy = [np.cos(trend[k]) * np.cos(plunge[k]) for k in range(np.size(trend))]

    # Return the coordinates as a tuple of numpy arrays
    return np.array(vx).T, np.array(vy).T, np.array(vz).T


def angleBtwVec(v1, v2):
    """
    Calculate the small angle between two directional oriented vectors.

    Parameters
    ----------
    v1 : array_like
        First vector in the form (x, y, z).
    v2 : array_like
        Second vector in the form (x, y, z).

    Returns
    -------
    float
        Angle between the two vectors in degrees.
    """

    if len(np.shape(v1)) > 1:
        v1n = np.array([j / np.linalg.norm(j) for j in v1])
        v2n = np.array([k / np.linalg.norm(k) for k in v2])
        dotv = np.array(
            np.round([np.dot(v1n[l], v2n[l].T) for l in range(len(np.shape(v1)))], 5)
        )
        angle = np.array([math.acos(dt) for dt in dotv])
    else:
        v1n = v1 / np.linalg.norm(v1)
        v2n = v2 / np.linalg.norm(v2)
        angle = math.acos(np.round(np.dot(v1n, v2n.T), 5))

    return angle / np.pi * 180


def azimRot(dip, trend, plunge):
    """Azimuth according to rotation axis and dip.

    Calculates the azimuth of a plane with a certain dip
    that goes through a rotation axis defined with trend and plunge angles.

    Parameters
    ----------
    dip : float
        Angles of plane dip (angles in degrees).
    trend : float
        Angles of trend (plunge azimuth) of the rotational axis.
    plunge : float
        Plunge of the rotational axis (angles in degrees).

    Returns
    -------
    float
        Angles of plane dip azimuths (angles in degrees) that pass through the axis.
    """

    trend = trend / 180 * np.pi
    dip = dip / 180 * np.pi
    plunge = plunge / 180 * np.pi
    azim = trend - np.pi / 2 + math.asin(math.tan(plunge) / math.tan(dip))

    return azim / np.pi * 180


def dipRot(azim, trend, plunge):
    """Dip According To Rotation Axis And Azimuth.

    Calculates the dip of a plane with a certain dip azimuth
    that passes through a rotation axis defined with trend and plunge angles.

    Parameters
    ----------
    azim : float
        Angles of plane dip azimuth (angles in degrees) (no strike).
    trend : float
        Angles of trend (plunge azimuth) of the rotational axis.
    plunge : float
        Plunge of the rotational axis (angles in degrees).

    Returns
    -------
    float
        Angles of plane dips (angles in degrees) that pass through the axis.
    """

    trend = trend / 180 * np.pi
    azim = azim / 180 * np.pi
    plunge = plunge / 180 * np.pi
    dip = math.atan(math.tan(plunge) / math.cos(azim - trend))

    return dip / np.pi * 180


def distance(P1, P2):
    """Calculate the shortest distance between two points in space.

    Parameters
    ----------
    P1 : list
        Coordinates of the first point [x, y, z].
    P2 : list
        Coordinates of the second point [x, y, z].

    Returns
    -------
    float
        Distance between two points.
    """

    return ((P1[0] - P2[0]) ** 2 + (P1[1] - P2[1]) ** 2) ** 0.5


def greatCircle(azim, dip):
    """Extract coordinates of the great circle of a given plane orientation.

    Parameters
    ----------
    azim : float
        Angle of azimuth (dip azimuth) of the plane.
    dip : float
        Dip angle of the plane.

    Returns
    -------
    list
        Coordinates of great circle points [x_stereo, y_stereo].
    """

    dip = np.array(dip) * np.pi / 180
    azim = -np.array(azim) * np.pi / 180 + np.pi / 2

    N = 51
    psi = np.linspace(0, np.pi, N)

    radint = math.tan(dip) * np.array([math.sin(ps) for ps in psi])
    radip = np.array([math.atan(radi) for radi in radint])
    rproj = np.array([math.tan((np.pi / 2 - radp) / 2) for radp in radip])
    x1 = rproj * np.array([math.sin(ps) for ps in psi])
    y1 = rproj * np.array([math.cos(ps) for ps in psi])

    x_stereo = np.array(
        [x1[k] * math.cos(azim) + y1[k] * math.sin(azim) for k in range(np.size(x1))]
    )
    y_stereo = np.array(
        [x1[k] * math.sin(azim) - y1[k] * math.cos(azim) for k in range(np.size(x1))]
    )

    return [x_stereo, y_stereo]


def neighborPoints(j, rows, columns, connectivity):
    """Define array containing all neighbor indexes of a certain point in a matrix.

    Parameters
    ----------
    j : int
        Index in the matrix of the point being analyzed (flattened matrix).
    rows : int
        Number of rows in the matrix.
    columns : int
        Number of columns in the matrix.
    connectivity : str
        Type of neighbor connectivity (8-connectivity or 4-connectivity).

    Returns
    -------
    numpy.ndarray
        Array with indexes of neighbors.
    """

    lowerleft = (columns * rows) - columns
    lowerright = columns * rows - 1

    if connectivity == 4:
        neigh = np.array([j + 1, j - 1, j - columns, j + columns])

        if np.remainder(j + 1, columns) == 0:
            neigh = np.array([j - 1, j - columns, j + columns])
        if np.remainder(j, columns) == 0:
            neigh = np.array([j + 1, j - columns, j + columns])
        if j < columns:
            neigh = np.array([j - 1, j + 1, j + columns])
        if j > lowerleft:
            neigh = np.array([j - 1, j + 1, j - columns])
        if j == 0:
            neigh = np.array([j + 1, j + columns])
        if j == columns - 1:
            neigh = np.array([j - 1, j + columns])
        if j == lowerleft:
            neigh = np.array([j + 1, j - columns])
        if j == lowerright:
            neigh = np.array([j - 1, j - columns])

    if connectivity == 8:
        neigh = np.array(
            [
                j + 1,
                j - 1,
                j - columns,
                j + columns,
                j - columns + 1,
                j - columns - 1,
                j + columns + 1,
                j + columns - 1,
            ]
        )

        if np.remainder(j + 1, columns) == 0:
            neigh = np.array(
                [j - 1, j - columns, j + columns, j - columns - 1, j + columns - 1]
            )
        if np.remainder(j, columns) == 0:
            neigh = np.array(
                [j + 1, j - columns, j + columns, j - columns + 1, j + columns + 1]
            )
        if j < columns:
            neigh = np.array(
                [j - 1, j + 1, j + columns, j + columns + 1, j + columns - 1]
            )
        if j > lowerleft:
            neigh = np.array(
                [j - 1, j + 1, j - columns, j - columns + 1, j - columns - 1]
            )
        if j == 0:
            neigh = np.array([j + 1, j + columns, j + columns + 1])
        if j == columns - 1:
            neigh = np.array([j - 1, j + columns, j + columns - 1])
        if j == lowerleft:
            neigh = np.array([j + 1, j - columns, j - columns + 1])
        if j == lowerright:
            neigh = np.array([j - 1, j - columns, j - columns - 1])

    return neigh


def neighborPointsD(j, rows, columns, dp):
    """Define array containing the indexes surrounding the index j at a given
       distance dp in a matrix [rows,columns] (8-connectivity).

    Parameters
    ----------
    rows : int
        Number of rows in the matrix.
    columns : int
        Number of columns in the matrix.
    j : int
        Index in matrix of the point being analyzed (flattened matrix).
    dp : int
        Pixel number at the wanted distance (1 being the closest neighbor possible).

    Returns
    -------
    numpy.ndarray
        Array with indexes of neighbors.
    """

    neigh = np.array([j + dp, j - dp, j - dp * columns, j + dp * columns])
    for i in range(1, dp + 1):
        neigh2 = np.array(
            [
                j - dp * columns + i,
                j - dp * columns - i,
                j + dp * columns + i,
                j + dp * columns - i,
            ]
        )
        neigh = np.concatenate((neigh, neigh2))
    for i in range(1, dp + 1):
        neigh2 = np.array(
            [
                j - (dp - i) * columns + dp,
                j - (dp - i) * columns - dp,
                j + (dp - i) * columns + dp,
                j + (dp - i) * columns - dp,
            ]
        )
        neigh = np.concatenate((neigh, neigh2))

    remj = np.remainder(j, columns)
    remn = [np.remainder(n, columns) for n in neigh]
    neigh = neigh.compress((remn <= remj + dp).flat)
    remn = [np.remainder(n, columns) for n in neigh]
    neigh = neigh.compress((remn >= remj - dp).flat)
    neigh = neigh.compress((neigh < rows * columns).flat)
    neigh = neigh.compress((neigh >= 0).flat)
    neigh = np.unique(neigh)

    return neigh


def nmbNeighbors(binaryimage):
    """Extract a matrix (of size of the binary image) that contains the
       number of positive neighbors for each pixel of the binary image.

    Parameters
    ----------
    binaryimage : numpy.ndarray
        Matrix of zeros and ones (or TRUE's and FALSE's).

    Returns
    -------
    numpy.ndarray
        Matrix with the number of positive neighbors.
    """

    bi = binaryimage.flatten()
    i_true = (bi == 1).nonzero()[0]
    nmbN = np.zeros(np.shape(bi))

    for i in i_true:
        neigh = neighborPoints(i, np.shape(binaryimage)[0], np.shape(binaryimage)[1], 8)
        nmbN[i] = np.size((bi[neigh] == 1).nonzero())

    nmbN_mat = np.reshape(nmbN, np.shape(binaryimage))
    return nmbN_mat


def normal2angle(normal):
    """Calculates the orientation (in dip azimuth and dip) of the plane
    according to its normal (in x, y, z).

    Parameters
    ----------
    normal : numpy.ndarray
        Normal vector: n = [x, y, z].

    Returns
    -------
    Tuple[float, float]
        Azimuth (dip azimuth) and dip of the plane expressed in angles.
    """

    if np.ndim(normal) > 1:
        normal = normal[0]
    if normal[2] < 0:
        normal = normal * np.array([-1, -1, -1])
    if normal[1] == 0:
        normal[1] = 0.000001

    zvector = np.array([0, 0, 1])
    dip = math.atan2(np.linalg.norm(np.cross(normal, zvector)), np.dot(normal, zvector))
    dip = dip / np.pi * 180
    azim = math.atan(abs(normal[0]) / abs(normal[1]))
    azim = azim / np.pi * 180

    if normal[0] > 0 and normal[1] <= 0:
        azim = 180 - azim

    if normal[0] <= 0 and normal[1] < 0:
        azim = 180 + azim

    if normal[0] <= 0 and normal[1] > 0:
        azim = 360 - azim

    return [azim, dip]


def plungeRot(azim, dip, trend):
    """Calculates the plunge of an axis with a given trend and through which a
    plane with a given orientation goes through.

    Parameters
    ----------
    azim : float
        Angles of plane orientation (angles degree °).
    dip : float
        Angles of plane dip (angles degree °).
    trend : float
        Angles of trend (plunge azimuth) (angles degree °).

    Returns
    -------
    float
        Angles of axis plunge (angles degree °) that hosts the given plane.
    """

    trend = trend / 180 * np.pi
    azim = azim / 180 * np.pi
    dip = dip / 180 * np.pi
    plunge = math.atan(math.tan(dip) * math.cos(azim - trend))

    return plunge / np.pi * 180


def sortLine(ind, matSize):
    """Sorts a vector of points to connect them into a line and attribute
    an order (2D only).

    Parameters
    ----------
    ind : array_like
        Indexes of points in a matrix.
    matSize : tuple
        Size/shape of matrix.

    Returns
    -------
    newi : ndarray
        Newly sorted/ordered indexes.
    """

    matx = np.arange(0, matSize[1])
    maty = np.arange(0, matSize[0])
    [x, y] = np.meshgrid(matx, maty)
    x = x.flatten()
    y = y.flatten()

    mat = np.zeros(np.shape(x))
    mat[ind] = 1
    mat = np.reshape(mat, (matSize[0], matSize[1]))

    newi = np.zeros(np.shape(ind))
    start = []
    for i in ind:
        neigh = neighborPoints(i, matSize[0], matSize[1], 8)
        npos = np.intersect1d(neigh, ind)

        if np.size(npos) == 1:
            start = i
            break

    if np.size(start) == 0:
        start = ind[0]
        neigh = neighborPoints(start, matSize[0], matSize[1], 8)
        npos = np.intersect1d(neigh, ind)
        newi[-1] = npos[0]
        ind = np.array(ind).compress((np.array(ind) != npos[0]).flat)

    count = 0
    newi[count] = start
    count = count + 1
    ind = np.array(ind).compress((np.array(ind) != start).flat)

    while np.size(ind) > 0:
        d = [distance([x[start], y[start]], [x[ii], y[ii]]) for ii in ind]
        dimin = [di for di in range(np.size(d)) if d[di] == np.min(d)]
        if len(dimin) > 1:
            for di in dimin:
                neigh = neighborPointsD(ind[di], matSize[0], matSize[1], 8)
                npos = np.intersect1d(neigh, ind)
                di_neigh = np.array(dimin).compress((np.array(dimin) != di).flat)
                npos = np.intersect1d(neigh, ind[di_neigh])
                if len(npos) == 0:
                    dimin = di
                    break

        newi[count] = ind[dimin]
        count = count + 1
        start = ind[dimin]
        ind = np.delete(ind, dimin)
    return newi


def stereoLine(trend, plunge):
    """Extracts coordinates of a projected line on a stereonet.

    Parameters
    ----------
    trend : float
        Angle of trend (plunge azimuth) of the line.
    plunge : float
        Plunge of the line.

    Returns
    -------
    [x_stereo, y_stereo] : list
        Coordinates of point(s) in a stereonet.
    """

    trend = np.array(trend / 180 * np.pi)
    plunge = np.array(plunge / 180 * np.pi)

    radius = np.tan(np.pi * (np.pi / 2 - plunge) / (2 * np.pi))
    x_stereo = np.sin(trend) * radius
    y_stereo = np.cos(trend) * radius

    if trend.compress((trend > np.pi / 2).flat) and trend.compress(
        (trend <= np.pi).flat
    ):
        trend = np.pi - trend
        x_stereo = np.sin(trend) * radius
        y_stereo = -np.cos(trend) * radius
    if trend.compress((trend > np.pi).flat) and trend.compress(
        (trend <= np.pi / 2 * 3).flat
    ):
        trend = trend - np.pi
        x_stereo = -np.sin(trend) * radius
        y_stereo = -np.cos(trend) * radius
    if trend.compress((trend > np.pi / 2 * 3).flat):
        trend = 2 * np.pi - trend
        x_stereo = -np.sin(trend) * radius
        y_stereo = np.cos(trend) * radius

    return [x_stereo, y_stereo]


def vect2angle(v):
    """Extracts from vectors (vx, vy, vz) its axis directions in angles (geology style).

    Parameters
    ----------
    v : list
        Vector: v = [vx, vy, vz].

    Returns
    -------
    trend, plunge : tuple
        Angles of trend (plunge azimuth) and plunge of the vector.
    """

    if np.ndim(v) > 1:
        v = v[0]

    if v[2] > 0:
        v = v * np.array([-1, -1, -1])

    if v[1] == 0:
        trend = 90
    else:
        trend = abs(math.atan(v[0] / v[1]))
        trend = trend / np.pi * 180

    plunge = math.atan(abs(v[2]) / np.sqrt(v[0] ** 2 + v[1] ** 2))
    plunge = plunge / np.pi * 180

    if v[0] > 0 and v[1] <= 0:
        trend = 180 - trend

    if v[0] <= 0 and v[1] <= 0:
        trend = 180 + trend

    if v[0] <= 0 and v[1] > 0:
        trend = 360 - trend

    return trend, plunge


def Mbox(title, text, style):
    """Creates a message box.

    Parameters
    ----------
    title : str
        Title/name of the box.
    text : str
        Message text.
    style : int
        Style of text (bold, fontsize, ...).

    Returns
    -------
    None
        Message box appears.
    """
    if platform.system() == "Windows":
        import ctypes

        ctypes.windll.user32.MessageBoxW(0, text, title, style)
    else:  # Assuming other platforms support tkinter
        root = tk.Tk()
        root.withdraw()  # Hide the main window

        messagebox.showinfo(title, text)

        root.destroy()



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 31 12:20:00 2021
@author: Anna Rauch

This script contains a set of functions/methods that are useful to
display and visualize the TIE-results.
"""

import numpy as np
import math
import matplotlib.pyplot as plt
from untie import TIE_general as TIEgen
from untie import TIE_classes as TIEclass
from mayavi import mlab
import matplotlib.patches as mpatches


def bed2cmap(BEDrst: np.ndarray, legendfile: str, leg_labels: bool = False) -> tuple:
    """
    Color Information For Bedrock.

    Transforms the raw-bedrock raster in a scalar matrix and extracts based
    on a prepared textfile the color (rgb-values) of the litho-stratigraphic
    units present and re-structures it in a cmap-matrix. If the legendfile
    contains label information, you can also extract the legend string values.

    Parameters
    ----------
    BEDrst : numpy.ndarray
        Bedrock raster matrix.
    legendfile : str
        Tab-separated textfile [bedrock_ID, R, G, B, label].
    leg_labels : bool, optional
        Boolean indicating whether to extract labels (e.g., 'Schrattenkalk') or not.
        Default is False.

    Returns
    -------
    tuple
        cm : numpy.ndarray
            Scalar matrix.
        cmap : numpy.ndarray
            Colormap for bedrock information.
        formations : list
            List of labels for legend. If leg_labels==False, then formations corresponds to the original Bedrock ID.
    """
    kind = np.unique(BEDrst.flatten())
    kind = np.extract(np.isnan(kind) == False, kind)
    cm = np.zeros(np.size(BEDrst.flatten()))
    leg = np.loadtxt(legendfile, usecols=(0, 1, 2, 3))
    k_col = np.zeros(np.shape(kind))

    for k in range(len(kind)):
        i = (BEDrst.flatten() == kind[k]).nonzero()
        cm[i] = k + 1
        k_col[k] = (leg[:, 0] == kind[k]).nonzero()[0]

    cm = cm.reshape(np.shape(BEDrst))
    cm = np.fliplr(cm)
    cmap = np.zeros((255, 4))

    if any(np.isnan(BEDrst.flatten())):
        c_step = np.linspace(0, 255, np.size(k_col) + 2)
        c_step = np.round(c_step, 0).astype(int)
        cmap[0 : int(c_step[1]), 0:3] = np.ones((int(c_step[1]), 3)) * [230, 230, 230]

        for ci in range(1, len(c_step) - 1):
            ind1 = int(c_step[ci])
            ind2 = int(c_step[ci + 1])
            rgb_c = leg[int(k_col[ci - 1]), 1:4]
            cmap[ind1:ind2, 0:3] = np.ones((ind2 - ind1, 3)) * rgb_c
    else:
        c_step = np.linspace(0, 255, np.size(k_col) + 1)
        c_step = np.round(c_step, 0).astype(int)

        for ci in range(len(c_step) - 1):
            ind1 = int(c_step[ci])
            ind2 = int(c_step[ci + 1])
            rgb_c = leg[int(k_col[ci]), 1:4]
            cmap[ind1:ind2, 0:3] = np.ones((ind2 - ind1, 3)) * rgb_c
    cmap[:, 3] = (np.ones((255, 1)) * 255).flatten()

    if leg_labels == True:
        f = open(legendfile)
        lines = f.readlines()
        lines = [lines[int(ki)] for ki in k_col]
        tokens_column_number = 4
        formations = []
        if any(np.isnan(BEDrst.flatten())):
            formations.append("Unconsolidated Deposit")
        for x in lines:
            formations.append(x.split("\t")[tokens_column_number][:-1])
        f.close()
    else:
        formations = [str(ki) for ki in kind]

    return cm, cmap, formations


def colorClassID(segment_OBJ: TIEclass.segment_OBJ) -> list:
    """
    Color Information Trace Classification.

    Extracts color code for TIE-classification.

    Parameters
    ----------
    segment_OBJ : TIEclass.segment_OBJ
        Segment object.

    Returns
    -------
    list
        Normalized RGB-value corresponding to TIE-class.
    """

    cmapblue = np.flipud(
        [
            [
                0,
                1,
                1,
            ],
            [0, 0.7, 1],
            [0, 0.4, 1],
            [0.3, 0, 0.8],
        ]
    )
    cmapred = np.flipud(
        [
            [0.98, 0.7, 0.9],
            [255 / 255, 100 / 255, 180 / 255],
            [216 / 255, 0, 113 / 255],
            [0.616, 0.106, 0.286],
        ]
    )
    classID = segment_OBJ.classID
    if len(segment_OBJ.ind_normal) < 50:
        c_code = [0.4, 0.4, 0.4]
    else:
        if classID > 0:
            c_code = cmapblue[classID - 1]
        else:
            c_code = cmapred[classID * -1 - 1]
    return c_code


def createCmap(rgbv: np.ndarray, n: int) -> np.ndarray:
    """
    Create Cmap.

    Creates a continuous color range (n values) from one rgb value (rgbv) to the other(s).

    Parameters
    ----------
    rgbv : numpy.ndarray
        Vector containing rgb-values of endmembers.
    n : int
        Resolution of cmap (number of values in cmap).

    Returns
    -------
    numpy.ndarray
        Newly created cmap.
    """

    cmap = np.zeros([n, 3])
    nint = int(np.round(n / (len(rgbv) - 1), 0))
    cint = np.concatenate((np.arange(0, n, nint), np.array(n)), axis=None)
    if len(rgbv) == 1:
        TIEgen.Mbox(
            "SORRY", "at least two rgb values are necessary to crate a colormap", 1
        )
    else:
        for j in range(len(rgbv) - 1):
            cint1 = cint[j]
            cint2 = cint[j + 1]
            dif = cint2 - cint1
            cstp = np.array(
                [
                    (rgbv[j + 1, 0] - rgbv[j, 0]) / (dif - 1),
                    (rgbv[j + 1, 1] - rgbv[j, 1]) / (dif - 1),
                    (rgbv[j + 1, 2] - rgbv[j, 2]) / (dif - 1),
                ]
            )
            if cstp[0] != 0:
                cmap1 = np.linspace(rgbv[j, 0], rgbv[j + 1, 0], dif).reshape((dif, 1))
            else:
                cmap1 = np.ones((dif, 1)) * rgbv[j, 0]

            if cstp[1] != 0:
                cmap2 = np.linspace(rgbv[j, 1], rgbv[j + 1, 1], dif).reshape((dif, 1))
            else:
                cmap2 = np.ones((dif, 1)) * rgbv[j, 1]

            if cstp[2] != 0:
                cmap3 = np.linspace(rgbv[j, 2], rgbv[j + 1, 2], dif).reshape((dif, 1))
            else:
                cmap3 = np.ones((dif, 1)) * rgbv[j, 2]
            cmap[cint[j] : cint[j + 1], :] = np.concatenate(
                (cmap1, cmap2, cmap3), axis=1
            )
    return cmap


def showOrientBars(traces, vx, vy, vz, stereo=False):
    """
    Show orientation bars along traces in 3D or stereo.

    Parameters
    ----------
    traces : list of trace_OBJ (TIE_classes.trace_OBJ)
        List of trace objects.
    vx, vy, vz : numpy.ndarray
        Flattened x-, y-, and z-coordinate matrices.
    stereo : bool, optional
        Boolean indicating whether to display the bars in stereo.
        If True, the bar length is stereographically projected on the map.
        Default is False.

    Returns
    -------
    bar : mayavi.modules.surface.Surface
        Handle to the plot.
    """

    for T in traces:
        index = T.index
        or_bar = T.orientbar
        cs = vx[0] - vx[1]
        barlength = 200 / cs

        for m in range(0, len(or_bar), int(20 / cs)):
            if stereo == True:
                tr, pl = TIEgen.vect2angle(or_bar[m])
                x1, y1 = TIEgen.stereoLine(tr, pl)
                bars = mlab.plot3d(
                    [vx[int(index[m])], vx[int(index[m])] + x1 * barlength],
                    [vy[int(index[m])], vy[int(index[m])] + y1 * barlength],
                    [vz[int(index[m])], vz[int(index[m])]],
                    tube_radius=3,
                    color=(0, 0, 0),
                )
            else:
                bars = mlab.plot3d(
                    [vx[int(index[m])], vx[int(index[m])] + or_bar[m][0] * barlength],
                    [vy[int(index[m])], vy[int(index[m])] + or_bar[m][1] * barlength],
                    [vz[int(index[m])], vz[int(index[m])] + or_bar[m][2] * barlength],
                    tube_radius=3,
                    color=(0, 0, 0),
                )
    return bars


def showOverview3D(
    mx,
    my,
    mz,
    mc,
    cmap=[],
    BTraces=[],
    FTraces=[],
    ATraces=[],
    WithlabelsB=True,
    WithlabelsF=True,
    WithlabelsA=True,
    Textsize=30,
):
    """Show Geological Map In 3d (No Tie)

    Displays the geological map in 3D, as well as the different trace sets and
    their ID (yet, no TIE classification).

    Parameters
    ----------
    mx, my, mz : array-like
        Coordinate matrices of the analyzed extent.
    mc : array-like
        Scalar field of bedrock id's.
    cmap : list, optional
        Bedrock colormap. Default is Mayavi's default cmap.
    BTraces : list, optional
        List of trace_OBJ of bedrock interface traces. Default is an empty list.
    FTraces : list, optional
        List of trace_OBJ of faults (tectonic boundaries). Default is an empty list.
    ATraces : list, optional
        List of trace_OBJ of axial traces. Default is an empty list.
    Textsize : int, optional
        Text size of trace labels. Default is 30.
    WithlabelsB : bool, optional
        Boolean indicating whether to add trace ID on the figure for bedrock traces.
        Default is True.
    WithlabelsF : bool, optional
        Boolean indicating whether to add trace ID on the figure for fault traces.
        Default is True.
    WithlabelsA : bool, optional
        Boolean indicating whether to add trace ID on the figure for axial traces.
        Default is True.

    Returns
    -------
    fig : mayavi.mlab.Figure
        Handle to the figure.
    """

    fig = mlab.figure(1, fgcolor=(1, 1, 1), bgcolor=(0.9, 0.9, 0.9))
    mesh = mlab.mesh(mx, my, mz, scalars=mc)
    if len(cmap) > 0:
        mesh.module_manager.scalar_lut_manager.lut.table = cmap

    vx = np.fliplr(mx).flatten()
    vy = my.flatten()
    vz = np.fliplr(mz).flatten()

    for tr in BTraces:
        mlab.plot3d(
            vx[tr.index.astype(int)],
            vy[tr.index.astype(int)],
            vz[tr.index.astype(int)],
            color=(0, 0, 0.8),
            tube_radius=2,
            name="bedrock trace " + str(tr.id),
        )
        if WithlabelsB:
            mlab.text3d(
                vx[tr.index.astype(int)[4]],
                vy[tr.index.astype(int)[4]],
                vz[tr.index.astype(int)[4]] + 70,
                str(tr.id),
                color=(0, 0, 0.8),
                scale=Textsize,
            )

    for fl in FTraces:
        mlab.plot3d(
            vx[fl.index.astype(int)],
            vy[fl.index.astype(int)],
            vz[fl.index.astype(int)],
            color=(1, 0, 0),
            tube_radius=2,
            name="fault trace " + str(fl.id),
        )
        if WithlabelsF:
            mlab.text3d(
                vx[fl.index.astype(int)[4]],
                vy[fl.index.astype(int)[4]],
                vz[fl.index.astype(int)[4]] + 70,
                str(fl.id),
                color=(1, 0, 0),
                scale=Textsize,
            )

    for ai in ATraces:
        mlab.plot3d(
            vx[ai.index.astype(int)],
            vy[ai.index.astype(int)],
            vz[ai.index.astype(int)],
            color=(1, 0, 0),
            tube_radius=1,
            name="axial trace " + str(ai.id),
        )
        if WithlabelsA:
            mlab.text3d(
                vx[ai.index.astype(int)[4]],
                vy[ai.index.astype(int)[4]],
                vz[ai.index.astype(int)[4]] + 70,
                str(ai.id),
                color=(1, 1, 0),
                scale=Textsize,
            )

    mlab.draw()
    mlab.show()
    return fig


def showOverviewMap(
    mx,
    my,
    mc,
    cmap=[],
    mz=[],
    BTraces=[],
    FTraces=[],
    ATraces=[],
    leg_labels=[],
    WithSeg=True,
):
    """Show Geological Map In 2d As An Overview

    Displays the geological map in 2D (with legend), as well as different
    trace sets and their ID (yet, no TIE classification).

    Parameters
    ----------
    mx, my, mz : array-like
        Coordinate matrices of the analyzed extent.
    mc : array-like
        Scalar field of bedrock id's.
    cmap : list, optional
        Bedrock colormap. Default is Matplotlib's default cmap.
    BTraces : list, optional
        List of trace_OBJ of bedrock interface traces. Default is an empty list.
    FTraces : list, optional
        List of trace_OBJ of faults (tectonic boundaries). Default is an empty list.
    ATraces : list, optional
        List of trace_OBJ of axial traces. Default is an empty list.
    leg_labels : list, optional
        List of labels for bedrock units. Default is an empty list.
    WithSeg : bool, optional
        Boolean - show segmentation. Default is True.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Handle to the figure.
    ax : matplotlib.axes._axes.Axes
        Handle to the axes.
    """
    fig, ax = plt.subplots()
    ax.tick_params(labelsize=6)

    if len(cmap) > 0:
        from matplotlib.colors import ListedColormap

        cmap_plt = ListedColormap(cmap / 255)
    else:
        cmap_plt = "Greens"

    vx = np.fliplr(mx).flatten()
    vy = my.flatten()
    ax.imshow(np.flipud(mc), cmap=cmap_plt, extent=(vx[-1], vx[0], vy[0], vy[-1]))

    if len(BTraces) > 0:
        for tr in BTraces:
            tx = vx[tr.index.astype(int)]
            ty = vy[tr.index.astype(int)]

            ax.plot(tx, ty, c="b")
            txt_name = str(tr.id)
            ax.text(
                tx[2],
                ty[2],
                txt_name,
                fontsize=6,
                fontweight="bold",
                c="b",
                bbox=dict(boxstyle="circle,pad=0", fc=(1, 1, 1, 0.7), lw=0),
            )
            if WithSeg == True:
                for s in tr.Segment:
                    si1 = int(s.ind_normal[0])
                    si2 = int(s.ind_normal[-1])
                    ax.scatter(tx[si1], ty[si1], s=1, c="k", zorder=3)
                    ax.scatter(tx[si2], ty[si2], s=1, c="k", zorder=3)
                    if len(tr.Segment) > 1:
                        simid = int(
                            s.ind_normal[0] + np.round(np.size(s.ind_normal) / 2)
                        )
                        seg_name = "(" + str(s.id) + ")"
                        ax.text(
                            tx[simid],
                            ty[simid],
                            seg_name,
                            fontsize=4,
                            fontweight="normal",
                            c="b",
                            bbox=dict(boxstyle="circle,pad=0", fc=(1, 1, 1, 0.7), lw=0),
                        )
    if len(FTraces) > 0:
        for fl in FTraces:
            fx = vx[fl.index.astype(int)]
            fy = vy[fl.index.astype(int)]
            ax.plot(fx, fy, c="r")
            txt_name = str(fl.id)
            ax.text(
                fx[2],
                fy[2],
                txt_name,
                fontsize=6,
                fontweight="bold",
                c="r",
                bbox=dict(boxstyle="circle,pad=0", fc=(1, 1, 1, 0.7), lw=0),
            )
            if WithSeg == True:
                for f in fl.Segment:
                    si1 = int(f.ind_normal[0])
                    si2 = int(f.ind_normal[-1])
                    simid = int(f.ind_normal[0] + np.round(np.size(f.ind_normal) / 2))
                    ax.scatter(fx[si1], fy[si1], s=1, c="k", zorder=3)
                    ax.scatter(fx[si2], fy[si2], s=1, c="k", zorder=3)
                    if len(fl.Segment) > 1:
                        seg_name = "(" + str(f.id) + ")"
                        ax.text(
                            fx[simid],
                            fy[simid],
                            seg_name,
                            fontsize=4,
                            fontweight="normal",
                            c="r",
                            bbox=dict(boxstyle="circle,pad=0", fc=(1, 1, 1, 0.7), lw=0),
                        )
    if len(ATraces) > 0:
        for atr in ATraces:
            atx = vx[atr.index.astype(int)]
            aty = vy[atr.index.astype(int)]
            ax.plot(atx, aty, c="y", linewidth=0.5)
            txt_name = str(atr.id)
            ax.text(
                atx[2],
                aty[2],
                txt_name,
                fontsize=6,
                fontweight="bold",
                c="y",
                bbox=dict(boxstyle="circle,pad=0", fc=(1, 1, 1, 0.7), lw=0),
            )
            if WithSeg == True:
                for s in atr.Segment:
                    si1 = int(s.ind_normal[0])
                    si2 = int(s.ind_normal[-1])
                    simid = int(s.ind_normal[0] + np.round(np.size(s.ind_normal) / 2))
                    ax.scatter(atx[si1], aty[si1], s=1, c="k", zorder=3)
                    ax.scatter(atx[si2], aty[si2], s=1, c="k", zorder=3)
                    if len(atr.Segment) > 1:
                        seg_name = "(" + str(s.id) + ")"
                        ax.text(
                            atx[simid],
                            aty[simid],
                            seg_name,
                            fontsize=4,
                            fontweight="normal",
                            c="y",
                            bbox=dict(boxstyle="circle,pad=0", fc=(1, 1, 1, 0.7), lw=0),
                        )
    if len(mz) > 0:
        levmin = (np.min(mz) - np.min(mz) % 10) + 10
        levmax = np.max(mz) - np.min(mz) % 10
        ax.contour(
            mz,
            extent=(vx[-1], vx[0], vy[0], vy[-1]),
            levels=np.arange(levmin, levmax, 20),
            colors="k",
            linewidths=0.1,
        )

    if len(leg_labels) > 0:
        indexes = np.unique(cmap, return_index=True, axis=0)
        cmap_leg = [cmap[index] / 255 for index in sorted(indexes[1])]
        patches = [
            mpatches.Patch(color=cmap_leg[i], label=leg_labels[i])
            for i in range(len(leg_labels))
        ]
        plt.legend(
            handles=patches,
            bbox_to_anchor=(1.05, 1),
            loc=2,
            borderaxespad=0.0,
            fontsize=6,
        )

    return fig, ax


def showSignal(trace_OBJ):
    """Show Trace Signals

    Displays signals of a specific trace. First subplot: major signals -
    alpha (normal and reverse) and beta (normal and reverse) and their
    signal heights. Second subplot: signal of the orthogonal distance
    between chords forming a chord plane (normalized to the total length of
    the trace).

    Parameters
    ----------
    trace_OBJ : trace object (TIE_classes.trace_OBJ)
        The trace object for which signals are to be displayed.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Handle to the figure.
    """
    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
    for s in trace_OBJ.Segment:
        Alpha = [chord.alpha[0] for chord in s.Chords]
        Beta = [chdplane.beta[0] for chdplane in s.Chordplanes]
        Dist = [chdplane.dist_ratio[0] for chdplane in s.Chordplanes]
        AlphaR = [chord.alpha[1] for chord in s.Chords]
        BetaR = [chdplane.beta[1] for chdplane in s.Chordplanes]
        ha = s.sigheight[0]
        hb = s.sigheight[1]

        l = len(s.ind_normal)
        la = len(Alpha)
        lb = len(Beta)

        lbeg = 100 * (s.id - 1)
        lend = 100 * s.id

        if s.id == 1:
            ax1.plot(
                np.linspace(lbeg, lend, la),
                Alpha,
                "g-",
                label="alpha normal",
                linewidth=1.5,
            )
            ax1.plot(
                np.linspace(lbeg, lend, la),
                AlphaR,
                "g-.",
                label="alpha reverse",
                linewidth=1.5,
            )
            ax1.plot([lbeg, lend], [ha, ha], "g:", linewidth=0.5, label="ha")
            ax1.plot(
                np.linspace(lbeg, lend, lb),
                Beta,
                "y-",
                label="beta normal",
                linewidth=1.5,
            )
            ax1.plot(
                np.linspace(lbeg, lend, lb),
                BetaR,
                "y-.",
                label="beta reverse",
                linewidth=1.5,
            )
            ax1.plot([lbeg, lend], [hb, hb], "y:", linewidth=0.5, label="hb")
        else:
            ax1.plot(np.linspace(lbeg, lend, la), Alpha, "g-", linewidth=1.5)
            ax1.plot(np.linspace(lbeg, lend, la), AlphaR, "g-.", linewidth=1.5)
            ax1.plot([lbeg, lend], [ha, ha], "g:", linewidth=0.5)
            ax1.plot(np.linspace(lbeg, lend, lb), Beta, "y-", linewidth=1.5)
            ax1.plot(np.linspace(lbeg, lend, lb), BetaR, "y-.", linewidth=1.5)
            ax1.plot([lbeg, lend], [hb, hb], "y:", linewidth=0.5)

        ax1.plot([lbeg, lbeg], [0, 180], "k:", linewidth=0.5)
        ax1.plot([lend, lend], [0, 180], "k:", linewidth=0.5)

        tit = "signal trace n° " + str(trace_OBJ.id)
        ax1.set_title(tit)
        ax1.set_xlabel("j / k [% of pixel length (m / l)]")
        ax1.set_ylabel("angle degrees [°]")
        ax1.text(
            lend + 0.4,
            30,
            "Seg. " + str(s.id) + " (" + str(l) + "pix)",
            rotation="vertical",
            fontsize=8,
            bbox=dict(boxstyle="round4,pad=0.1", fc=(1, 1, 1, 1), lw=0),
        )
        ax1.axis([-3, lend + 3, -1, 182])

        if s.id == 1:
            ax2.plot(
                np.linspace(lbeg, lend, lb),
                np.array(Dist) * 100,
                "y-",
                linewidth=1.5,
                label="orthogonal distance btw chords",
            )
        else:
            ax2.plot(
                np.linspace(lbeg, lend, lb), np.array(Dist) * 100, "y-", linewidth=1.5
            )
        ax2.set_xlabel("k in % of trace length (l)")
        ax2.set_ylabel("d in % of trace length")

        ax2.plot([lbeg, lbeg], [0, 10], "k:", linewidth=0.5)
        ax2.plot([lend, lend], [0, 10], "k:", linewidth=0.5)
        ax2.plot([lbeg, lend], [2, 2], "k", linewidth=0.8)
        ax2.plot([lbeg, lend], [5, 5], "k", linewidth=0.8)
        ax2.plot([lbeg, lend], [10, 10], "k", linewidth=0.8)
        ax2.text(
            lend + 0.4,
            2,
            "Seg. " + str(s.id) + " (" + str(l) + "pix)",
            rotation="vertical",
            fontsize=8,
            bbox=dict(boxstyle="round4,pad=0.1", fc=(1, 1, 1, 1), lw=0),
        )

    ax1.legend(loc="upper left", fontsize="xx-small")
    ax2.legend(loc="upper left", fontsize="xx-small")

    ax2.text(0.4, 1.2, "chord planes have physical meaning", fontsize=6)
    ax2.text(0.4, 4.2, "only vague physical meaning", fontsize=6)
    ax2.text(0.4 + 0.2, 8.2, "no physical meaning", fontsize=6)
    plt.xticks(
        np.arange(0, len(trace_OBJ.Segment) * 100 + 100, 100),
        [0] + [100] * (len(trace_OBJ.Segment)),
    )
    ax2.axis([-3, lend + 3, 0, 11])

    plt.show
    return fig


def showSigStereo(trace_OBJ):
    """Show Trace Signals As Stereographic Projection

    Displays the evolution of the chords (1st subplot) and chord planes
    (2nd plot) of a specific trace by means of a stereographic projection.

    Parameters
    ----------
    trace_OBJ : trace object (TIE_classes.trace_OBJ)
        The trace object for which signals are to be displayed.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Handle to the figure.
    """
    ls = len(trace_OBJ.Segment)
    fig, axs = plt.subplots(ls, 2)
    fig.suptitle("Stereographic projection: trace n° " + str(trace_OBJ.id), fontsize=12)
    for s in trace_OBJ.Segment:
        Chords = s.Chords
        l = len(Chords)
        axChd = [chord.vec_trend[0] for chord in Chords]
        axPl = [chord.vec_plunge[0] for chord in Chords]
        ai = s.id - 1

        vec = np.array([TIEgen.stereoLine(axChd[i], axPl[i]) for i in range(l)])
        x_Ax = vec[:, 0]
        y_Ax = vec[:, 1]

        if l % 2 == 0:
            half = int(l / 2)
            halfpoint = int(half)
        else:
            half = int((l - 1) / 2)
            halfpoint = int(half + 1)

        rgbv = np.array([[1, 1, 0], [0.13, 0.54, 0.13], [0, 0, 0.8]])
        cmap = createCmap(rgbv, halfpoint)

        if ls == 1:
            ai_chd = 0
            ai_chdp = 1
        else:
            ai_chd = (ai, 0)
            ai_chdp = (ai, 1)

        axs[ai_chd] = stereoPlot(axs[ai_chd])
        for i in range(halfpoint):
            axs[ai_chd].scatter(x_Ax[i], y_Ax[i], s=25, c=[cmap[i, :]])
            axs[ai_chd].scatter(x_Ax[i + half], y_Ax[i + half], s=25, c=[cmap[i, :]])
        axs[ai_chd].text(-1, 1, "seg. " + str(s.id), fontsize=6)

        ChdPl = s.Chordplanes
        N = len(ChdPl)
        axs[ai_chdp] = stereoPlot(axs[ai_chdp])
        for i in range(N):
            azim = ChdPl[i].plane_orient[0]
            dip = ChdPl[i].plane_orient[1]
            [x, y] = TIEgen.greatCircle(azim, dip)

            axs[ai_chdp].plot(x, y, c=tuple(cmap[i, :]), linewidth=0.3)
            x_stereo, y_stereo = TIEgen.stereoLine(
                ChdPl[i].pole_orient[0], ChdPl[i].pole_orient[1]
            )
            axs[ai_chdp].scatter(x_stereo, y_stereo, s=25, c=[cmap[i, :]])
        axs[ai_chdp].text(-1, 1, "seg. " + str(s.id), fontsize=6)

    axs[ai_chd].set_xlabel("chord vector orientations", fontsize=8)
    axs[ai_chdp].set_xlabel("chord plane (gc) & pole orientations", fontsize=8)
    return fig


def showTIEmap(
    mx: np.ndarray,
    my: np.ndarray,
    mz: np.ndarray,
    mc: list = [],
    cmap: str = "",
    MainTrace_Set: list = [],
    AuxTrace_Set: list = [],
    ShowBars: bool = True,
) -> mlab.figure:
    """
    Show Tie-Results In 3d.

    Displays the TIE-results in 3D on top of the slightly transparent geological map.
    TIE-results: (1) trace segments are colored according to TIE-class, (2) orientation
    bars indicate chord plane orientations along the trace.

    Parameters
    ----------
    mx : numpy.ndarray
        Coordinate matrixes of analyzed extent.
    my : numpy.ndarray
        Coordinate matrixes of analyzed extent.
    mz : numpy.ndarray
        Coordinate matrixes of analyzed extent.
    mc : list, optional
        Scalar field of bedrock id's.
    cmap : str, optional
        Bedrock colormap. Default is mayavi's default cmap.
    MainTrace_Set : list, optional
        List of tie-analyzed trace objects.
    AuxTrace_Set : list, optional
        List of auxiliary trace objects that are to be neutrally displayed (e.g., fault traces).
    ShowBars : bool, optional
        Boolean indicating whether to show orientation bars or not. Default is True.

    Returns
    -------
    mayavi.modules.figure.Figure
        Handle to the figure.
    """

    if len(cmap) == 0:
        cmap = "Greens"

    fig = mlab.figure(2, fgcolor=(1, 1, 1), bgcolor=(0.9, 0.9, 0.9))
    mesh = mlab.mesh(mx, my, mz, scalars=mc, opacity=0.6)
    mesh.module_manager.scalar_lut_manager.lut.table = cmap

    vx = np.fliplr(mx).flatten()
    vy = my.flatten()
    vz = np.fliplr(mz).flatten()

    # analyzed trace set
    for tr in MainTrace_Set:
        tx = vx[tr.index.astype(int)]
        ty = vy[tr.index.astype(int)]
        tz = vz[tr.index.astype(int)]
        for s in tr.Segment:
            mlab.plot3d(
                tx[s.ind_normal],
                ty[s.ind_normal],
                tz[s.ind_normal],
                color=tuple(colorClassID(s)),
                tube_radius=5,
                name="analyzed trace " + str(tr.id),
            )
            mlab.text3d(
                tx[s.ind_normal[4]],
                ty[s.ind_normal[4]],
                tz[s.ind_normal[4]] + 70,
                str(tr.id),
                color=(0, 0, 0),
                scale=25,
            )

    # auxiliary trace set
    for fl in AuxTrace_Set:
        mlab.plot3d(
            vx[fl.index.astype(int)],
            vy[fl.index.astype(int)],
            vz[fl.index.astype(int)],
            color=(0, 0, 0),
            tube_radius=1.5,
            name="auxiliary trace " + str(fl.id),
        )

    if ShowBars == True:
        showOrientBars(MainTrace_Set, vx, vy, vz)

    mlab.draw()
    mlab.show()

    return fig


def sigHdiagram(trace_set, pth=[3, 9, 18, 90], txt=True, scale="lin"):
    """
    Show Signal Height Diagram.

    Displays a scatter plot with the signal height h_alpha as a function of the
    signal height h_beta for each trace in a trace set. The classification
    boundaries (pth) are displayed as background information. The dots are
    colored according to their TIE-classification.

    Parameters
    ----------
    trace_set : list of trace_OBJ (TIE_classes.trace_OBJ)
        List of trace objects.
    pth : list of int, optional
        Planarity threshold used to classify the traces.
        Default is [3, 9, 18, 90].
    txt : bool, optional
        Boolean indicating whether to display the trace's id or not.
        Default is True.
    scale : str, optional
        Scale for displaying the SH-diagram. "log" for logarithmic scale,
        "lin" for linear scale. Default is "lin".

    Returns
    -------
    fig : matplotlib.figure.Figure
        Handle to the figure.
    """

    x1 = [
        0.0001,
        0.0005,
        0.001,
        0.005,
        0.01,
        0.05,
        0.1,
        0.2,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
        0.8,
        0.9,
        1,
    ]
    x2 = [
        1.2,
        1.2,
        1.5,
        1.8,
        2,
        2.5,
        3,
        3.8,
        5,
        6,
        8,
        10,
        15,
        20,
        30,
        40,
        50,
        60,
        70,
        90,
        180,
    ]
    x0 = x1 + x2
    y0 = [180 / x0i for x0i in x0]

    fig, ax = plt.subplots()
    if scale == "log":
        ax.axis([3, 180, 3, 180])
        ax.set_yscale("log")
        ax.set_xscale("log")
    else:
        ax.axis([0, 180, 0, 180])
    ax.set_aspect("equal", adjustable="box")

    ax.plot(x0, y0, "k:", linewidth=0.7)
    ax.text(
        11,
        11,
        "p = 0°",
        fontsize=8,
        rotation=-45,
        bbox=dict(boxstyle="round4,pad=0.5", fc=(1, 1, 1, 1), lw=0),
    )
    ax.plot([0, 180], [0, 180], "k--", linewidth=1)
    ax.text(
        120,
        120,
        "s = 1",
        fontsize=8,
        rotation=45,
        bbox=dict(boxstyle="round4,pad=0.5", fc=(1, 1, 1, 1), lw=0),
    )

    ax.set_title("Signal-Height-Diagram")
    ax.set_xlabel("h_alpha")
    ax.set_ylabel("h_beta")

    for j in pth:
        x = np.array(x0) + j
        y = 180 / (x - j) + j
        ax.plot(x, y, "k", linewidth=1)
        ax.text(
            170,
            y[-1],
            str(j) + "°",
            fontsize=6,
            bbox=dict(boxstyle="round4,pad=0.1", fc=(1, 1, 1, 1), lw=0),
        )

    for tr in trace_set:
        sl = len(tr.Segment)
        for s in tr.Segment:
            ax.scatter(s.sigheight[0], s.sigheight[1], s=10, c=[colorClassID(s)])
            if txt == True:
                if sl == 1:
                    txt_name = str(tr.id)
                else:
                    txt_name = str(tr.id) + "(" + str(s.id) + ")"
                if scale == "log":
                    dec = s.sigheight[0] / 30
                else:
                    dec = 2

                ax.text(
                    s.sigheight[0] + dec,
                    s.sigheight[1],
                    txt_name,
                    fontsize=6,
                    bbox=dict(boxstyle="round4,pad=0", fc=(1, 1, 1, 0.6), lw=0),
                )
    plt.show
    return fig


def stereoPlot(ax, WithCircles="no"):
    """
    Stereoplot.

    Displays a stereonet ("Wulff - preservation of angles").

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Handle to the axes in which the stereonet should be displayed.
    with_circles : bool, optional
        Boolean indicating whether to show the great circles or only show major axes.
        Default is False.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Handle to the axes.
    """

    N = 50
    v = np.linspace(0, 2 * np.pi, N)
    cx = [math.cos(vi) for vi in v]
    cy = [math.sin(vi) for vi in v]
    xh = [-1, 1]
    yh = [0, 0]
    xv = [0, 0]
    yv = [-1, 1]
    lines = ax.plot(xh, yh, xv, yv, cx, cy, [0, 0], [1, 1.05])
    plt.setp(lines, color="k", linewidth=0.7)

    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.tick_params(axis="x", which="both", bottom=False, top=False, labelbottom=False)
    ax.tick_params(axis="y", which="both", right=False, left=False, labelleft=False)
    ax.text(-0.04, 1.07, "N")

    if WithCircles == "yes":
        vhf = np.linspace(0, np.pi, N)
        for i in range(1, 9):
            rdip = i * (np.pi / 18)
            rint = np.array(math.tan(rdip) * np.array([math.sin(vi) for vi in vhf]))
            radip = np.array([math.atan(rt) for rt in rint])
            rproj = np.array([math.tan((np.pi / 2 - rdp) / 2) for rdp in radip])
            x1 = rproj * np.array([math.sin(vi) for vi in vhf])
            x2 = rproj * np.array([(-math.sin(vi)) for vi in vhf])
            y = rproj * np.array([math.cos(vi) for vi in vhf])

            ax.plot(x1, y, ":k", x2, y, ":k", linewidth=0.4)
        for i in range(1, 9):
            alp = i * (np.pi / 18)
            xlim = math.sin(alp)
            x = np.arange(-xlim, xlim, 0.01)
            d = 1 / math.cos(alp)
            rd = d * math.sin(alp)
            y00 = rd**2 - np.array([xi**2 for xi in x])
            y0 = np.array([math.sqrt(yi) for yi in y00])
            y1 = np.array(d - np.array([yi for yi in y0]))
            y2 = np.array(-d + np.array([yi for yi in y0]))

            ax.plot(x, y1, ":k", x, y2, ":k", linewidth=0.4)
    ax.axis([-1, 1, -1, 1.15])
    ax.set_aspect("equal", "box")
    plt.draw()
    return ax



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 23 11:43:27 2021
@author: Anna Rauch

This script contains a set of object definitions (classes), which are the main
structure forms, in which the trace information is stored. It is a tree-based
structure, so that list of objects can be stored within a different object.
As an example: a trace_OBJ stores a list of segment_OBJ, each segment_OBJ
itself contains a list of chordPlane_OBJ...
"""


class chord_OBJ(object):
    def __init__(self, ID, vector, vec_trend, vec_plunge, alpha):
        self.id = ID
        self.vector = vector
        self.vec_trend = vec_trend
        self.vec_plunge = vec_plunge
        self.alpha = alpha


class chordPlane_OBJ(object):
    def __init__(
        self, ID, normal, plane_orient, pole_orient, beta, dist_meters, dist_ratio
    ):
        self.id = ID
        self.normal = normal
        self.plane_orient = plane_orient
        self.pole_orient = pole_orient
        self.beta = beta
        self.dist_meters = dist_meters
        self.dist_ratio = dist_ratio


class geolMap_OBJ(object):
    def __init__(self, name, mx, my, mz, BEDrst, BTraces, FTraces, ATraces, ORmeas):
        self.name = name
        self.mx = mx
        self.my = my
        self.mz = mz
        self.BEDrst = BEDrst
        self.BTraces = BTraces
        self.FTraces = FTraces
        self.ATraces = ATraces
        self.ORmeas = ORmeas


class segment_OBJ(object):
    def __init__(
        self, ID, delta, index_N, index_R, chords, chordplanes, classID, signalheight
    ):
        self.id = ID
        self.delta = delta
        self.ind_normal = index_N
        self.ind_reverse = index_R
        self.Chords = chords
        self.Chordplanes = chordplanes
        self.classID = classID
        self.sigheight = signalheight


class trace_OBJ(object):
    def __init__(self, ID, index, Type, Segment, convexityTH, orientbar):
        self.id = ID
        self.index = index
        self.type = Type
        self.Segment = Segment
        self.convexityTH = convexityTH
        self.orientbar = orientbar



# -*--- coding: utf-8 -*-
"""
General functions/methods that allow you 
to load the data and extract the traces in a way needed to perform a
TIE-analysis.
"""

import glob
import os
from pathlib import Path, PurePath
import tempfile

from typing import Dict, List, Tuple, Union

import geopandas as gpd
import numpy as np
import rasterio as rst
from untie import TIE_classes as TIEclass
from untie import TIE_general as TIEgen
from rasterio.mask import mask
from rasterio.merge import merge
from shapely.geometry import Polygon
from shapely.geometry import box


def adaptSHAPE2DEM(shapefile: gpd.GeoDataFrame, DEM: Dict) -> gpd.GeoDataFrame:
    """
    Adapt a shapefile's extent to match the extent given by a DEM.

    Parameters
    ----------
    shapefile : geopandas.GeoDataFrame
        Geopandas shapefile object.
    DEM : dict
        Dictionary containing DEM and coordinate data obtained with TIE_load.cropDEMextent.

    Returns
    -------
    geopandas.GeoDataFrame
        Clipped shapefile.

    Notes
    -----
    The DEM parameter should be a dictionary with the following keys:
        - 'meta': Metadata of the DEM.
        - 'z': DEM data.
    """

    DEMmeta = DEM["meta"]
    DEMtrans = DEMmeta["transform"]
    DEMz = DEM["z"]

    x = (DEMtrans[2], DEMtrans[2] + DEMtrans[0] * DEMz.shape[1])
    y = (DEMtrans[5], DEMtrans[5] + DEMtrans[4] * DEMz.shape[0])
    crs_shp = shapefile.crs
    df_ext = createExtentPLG(x, y, crs_shp)

    shapefile_clipped = gpd.overlay(shapefile, df_ext, how="intersection")

    return shapefile_clipped


def createExtentPLG(x, y, crs_shp):
    """Creates Extent Polygon.

    Creates a georeferenced polygon (geopandas) covering the extent defined
    with x- and y-coordinates.

    Parameters
    ----------
    x : list of float
        X-coordinates of the wanted polygon.

    y : list of float
        Y-coordinates of the wanted polygon.

    crs_shp : geopandas.crs.CRS
        Coordinate system of the shape (geopandas object).

    Returns
    -------
    geopandas.geodataframe.GeoDataFrame
        Geopandas object (shapefile) with the polygon extent.
    """

    extent = gpd.GeoSeries(
        Polygon([(x[0], y[0]), (x[0], y[1]), (x[1], y[1]), (x[1], y[0])])
    )
    extent = extent.set_crs(crs_shp)
    df_ext = gpd.GeoDataFrame({"geometry": extent})

    return df_ext


def cropDEMextent(geotif: rst.io.DatasetReader, shapefile: gpd.GeoDataFrame) -> dict:
    """Crop DEM with Shapefile.

    Crops the DEM extent according to a PLG shapefile (often an extent shapefile).

    Parameters
    ----------
    geotif : rasterio.io.DatasetReader
        Handle to an opened geotif (rasterio).

    shapefile : gpd.GeoDataFrame
        Handle to an opened shapefile (geopandas object).

    Returns
    -------
    dict
        Dictionary containing coordinate data.
        - 'z': Numpy array representing the cropped DEM.
        - 'x': Numpy array representing the x-coordinates.
        - 'y': Numpy array representing the y-coordinates.
        - 'meta': Metadata for the cropped DEM.
    """

    bbox = box(
        shapefile.total_bounds[0],
        shapefile.total_bounds[1],
        shapefile.total_bounds[2],
        shapefile.total_bounds[3],
    )

    gbbox = gpd.GeoDataFrame({"geometry": bbox}, index=[0], crs=shapefile.crs)

    def getFeatures(gdf):
        import json

        return [json.loads(gdf.to_json())["features"][0]["geometry"]]

    coords = getFeatures(gbbox)

    DEM, out_trans = mask(geotif, coords, crop=True)
    z = DEM[0]

    xend = out_trans[2] + out_trans[0] * z.shape[1] + 1
    x1 = out_trans[2] + 1
    y1 = out_trans[5] + out_trans[4] * z.shape[0] + 1
    yend = out_trans[5] + 1

    x = np.arange(x1, xend, out_trans[0])
    y = np.arange(y1, yend, out_trans[0])

    out_meta = geotif.meta.copy()
    out_meta.update(
        {
            "driver": "GTiff",
            "height": DEM.shape[0],
            "width": DEM.shape[1],
            "transform": out_trans,
        }
    )

    DEM = {"z": z, "x": x, "y": y, "meta": out_meta}
    return DEM


def extractTraces(TRmatrix, shape):
    """Extract Traces as Trace Objects.

    Extracts individual, sorted traces (as a trace_OBJ) based on a trace matrix.

    Parameters
    ----------
    TRmatrix : numpy.ndarray
        Matrix containing the traces (binary image), in the form of its trace matrix.

    shape : str
        'L' for polylines (e.g., faults), 'PLG' for traces extracted from polygons
        (e.g., bedrock interface traces).

    Returns
    -------
    list
        List of trace objects (trace_OBJ) defined in TIE_classes.
    """

    kind = np.unique(np.extract(np.isnan(TRmatrix) == False, TRmatrix))
    TRACE = []

    if np.size(kind) > 0:
        n = 1
        for r in kind:
            tr = np.zeros(np.shape(TRmatrix.flatten()))
            tr[(TRmatrix.flatten() == r)] = 1
            tr_mat = np.reshape(tr, np.shape(TRmatrix))

            if shape == "L":
                nmbN = TIEgen.nmbNeighbors(tr_mat).flatten()
                BP = (nmbN > 2).nonzero()
                tr[BP] = 0
                tr_mat = np.reshape(tr, np.shape(TRmatrix))

            from skimage import measure

            CC = measure.label(tr_mat)
            CCmat = CC.flatten()
            ntr = np.unique(CC)[1:]

            for k in ntr:
                index = (CCmat == k).nonzero()[0]
                if np.size(index) > 5:
                    index = TIEgen.sortLine(index, np.shape(tr_mat))
                    i_ind = np.arange(0, np.size(index)).astype(int)
                    seg = TIEclass.segment_OBJ(
                        1, [], i_ind, i_ind[::-1], [], [], [], []
                    )
                    TRACE.append(TIEclass.trace_OBJ(n, index, r, [seg], None, []))
                    n = n + 1
    return TRACE


def findNeighType(
    TRACES: List[TIEclass.trace_OBJ], BEDrst: np.ndarray
) -> List[TIEclass.trace_OBJ]:
    """
    Extracts the TWO bedrock types stored in BEDrst, which form a bedrock interface trace.

    Parameters
    ----------
    TRACES : List[TIEclass.trace_OBJ]
        List of trace_OBJ storing trace information of bedrock interface traces.
    BEDrst : numpy.ndarray
        Bedrock matrix (rasterized version of bedrock shapefile).

    Returns
    -------
    List[TIEclass.trace_OBJ]
        List of trace_OBJ with added SECOND bedrock type defining a boundary type.
    """

    for T in TRACES:
        mid = int(round(np.size(T.index) / 2))
        ind = int(T.index[mid])
        neigh = TIEgen.neighborPoints(ind, np.shape(BEDrst)[0], np.shape(BEDrst)[1], 4)
        BEDfl = BEDrst.flatten()
        rtype = BEDfl[ind]

        for k in neigh:
            if BEDfl[k] != rtype and np.isnan(BEDfl[k]) == False:
                ntype = BEDfl[k]
                break
            else:
                ntype = float("NaN")  # TODO: np.nan
        T.type = [rtype, ntype]

    return TRACES


def identifyTRACE(BEDrst: np.ndarray, FAULTS: List[TIEclass.trace_OBJ]) -> np.ndarray:
    """
    Identifies traces in a bedrock matrix based on the bedrock distinction and the fault traces.

    Parameters
    ----------
    BEDrst : numpy.ndarray
        Matrix with Bedrock data (see loadBedrock and rasterizeBedrock).
    FAULTS : List[TIEclass.trace_OBJ]
        List of trace_OBJ containing FAULT information.

    Returns
    -------
    numpy.ndarray
        Matrix (same size as BEDrst) with trace information.
    """

    TRmatrix = np.zeros(np.shape(BEDrst)).flatten()
    BEDwork = BEDrst.flatten()

    if FAULTS:
        faultind = np.concatenate(
            [FAULTS[fi].index for fi in range(np.shape(FAULTS)[0])]
        )
        faultind = [int(np.unique(fi)) for fi in faultind]
        fneighind = []
        for i in faultind:
            neigh = TIEgen.neighborPoints(
                i, np.shape(BEDrst)[0], np.shape(BEDrst)[1], 8
            )
            fneighind = np.concatenate((fneighind, neigh))
        fneighind = np.unique(fneighind)
        fneighind = [int(np.unique(fni)) for fni in fneighind]
    else:
        faultind = []
        fneighind = []

    BEDwork[faultind + fneighind] = float("NaN")
    indpos = (np.isnan(BEDwork) == False).nonzero()
    for i in indpos[0]:
        neigh = TIEgen.neighborPoints(i, np.shape(BEDrst)[0], np.shape(BEDrst)[1], 4)
        neigh = neigh.compress((np.isnan(BEDwork[neigh]) == False).flat)
        if any(BEDwork[neigh] > BEDwork[i]):
            TRmatrix[i] = BEDwork[i]

    TRmatrix[(TRmatrix == 0).nonzero()] = float("NaN")
    TRmatrix = np.reshape(TRmatrix, np.shape(BEDrst))

    return TRmatrix


def loadGeocover(
    sheet: int, geoc_path: str, language: str = "de"
) -> Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """
    Loads Geocover Data (all geology data) according to the Geocover structure.

    Parameters
    ----------
    sheet : int
        Identification number of LK sheet (Landeskarte 1.25'000).
    geoc_path : str
        Path to Geocover folder.
    language : str, optional
        Language for the version ("de" for German, "fr" for French), default is "de".

    Returns
    -------
    Tuple[geopandas.GeoDataFrame, geopandas.GeoDataFrame, geopandas.GeoDataFrame]
        Geopandas handles to shapefiles of Bedrock_PLG.shp, Tectonic_Boundaries_L.shp, and Planar_Structures_PT.shp.
    """

    path_whole = geoc_path + "/GC-V-V2-" + str(sheet) + "/Data/SHP/" + language
    BED_path = Path(path_whole + "/Bedrock_PLG.shp")
    TEC_path = Path(path_whole + "/Tectonic_Boundaries_L.shp")
    OM_path = Path(path_whole + "/Planar_Structures_PT.shp")

    BED = gpd.read_file(BED_path)
    TEC = gpd.read_file(TEC_path)
    OM = gpd.read_file(OM_path)

    return BED, TEC, OM


def loadLKdem(sheet: int, path_alti3D_folder: str) -> Union[rst.DatasetReader, None]:
    """
    Loads DEM (swissALTI3D), gluing the different patches together.

    Parameters
    ----------
    sheet : int
        Identification number of LK sheet (Landeskarte 1:25'000).
    path_alti3D_folder : str
        Path to swissALTI3D folder.

    Returns
    -------
    Union[rasterio.DatasetReader, None]
        Rasterio handle to DEM of the whole LK sheet (Landeskarte 1:25'000), or None if loading fails.
    """

    search_criteria = "/swiss_" + str(sheet) + "_*.tif"
    dem_fps = glob.glob(path_alti3D_folder + search_criteria)

    dem_mosaic = []
    for d in dem_fps:
        dem1 = rst.open(d)
        dem_mosaic.append(dem1)

    mosaic, out_trans = merge(dem_mosaic)
    out_meta = dem1.meta.copy()
    out_meta.update(
        {
            "driver": "GTiff",
            "height": mosaic.shape[1],
            "width": mosaic.shape[2],
            "transform": out_trans,
        }
    )

    # FIXME: No static paths inside script body
    tmp_dir = Path(tempfile.mkdtemp())
    if not os.path.isdir(tmp_dir):
        os.mkdirs(tmp_dir)
    # out_fp = Path("D:/TIE/python/temp/swiss_" + str(sheet) + "_merged.tif")
    out_fp = PurePath(tmp_dir, str(sheet) + "_merged.tif")
    with rst.open(out_fp, "w", **out_meta) as dest:
        dest.write(mosaic)

    merged = rst.open(out_fp)

    return merged


def rasterizeSHP(shp, attribute, DEM):
    """Rasterizes Shapefile.

    Rasterizes a shapefile according to a specific attribute field value.

    Parameters
    ----------
    shp : geopandas.geodataframe.GeoDataFrame
        Opened geopandas handle to a shapefile.

    attribute : str
        Attribute in shapefile that will be used to distinguish between different types
        of tectonic boundaries or litho-stratigraphic units. The attribute value must be a number.

    DEM : dict
        Dictionary containing coordinates (x, y, and z) of the analyzed zone (see crop2DEMextent).

    Returns
    -------
    numpy.ndarray
        Raster matrix.
    """

    KIND = getattr(shp, attribute)
    shp["KIND_NUM"] = KIND.astype(int)

    res = DEM["meta"]["transform"][0]
    mat = np.zeros(np.shape(DEM["z"]))
    matx, maty = np.meshgrid(DEM["x"], DEM["y"])

    from geocube.api.core import make_geocube

    cube = make_geocube(shp, measurements=["KIND_NUM"], resolution=(res, -res))

    xc = cube.KIND_NUM["x"].values
    yc = cube.KIND_NUM["y"].values

    xci = np.intersect1d(DEM["x"][::-1], xc, return_indices=True)[1]
    yci = np.intersect1d(DEM["y"], yc, return_indices=True)[1]
    matxi, matyi = np.meshgrid(xci, yci)
    matxi = np.fliplr(matxi)
    matv = cube.KIND_NUM.values.flatten()
    mat[matyi.flatten(), matxi.flatten()] = matv

    return mat



import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="untie",
    version=read("VERSION").strip(),
    author="Anna Rauch",
    description=("Trace Information Extraction (TIE) library"),
    author_email="contact@infogrip.ch",
    maintainer="Marc Monnerat",
    maintainer_email="marc.monnerat@swisstopo.ch",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    packages=["untie"],
    license_files=("LICENSE.md",),
    install_requires=[
        "shapely>=2.0.0",
        "geocube",
        "numpy",
        "geopandas",
        "mayavi",
        "matplotlib",
        "rasterio",
        "scipy",
        "scikit-image",
    ],
    package_data={
        "untie": [
            "src",
        ],
    },
    
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: GIS",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 23 09:19:29 2021
@author: Anna Rauch

This script contains a set of general functions/methods that perform the
core-steps of the TIE-method (i.e. the actual TIE-steps)
"""
import copy
from typing import Dict, List, Tuple, Union

import numpy as np
from untie import TIE_classes as TIEclass
from untie import TIE_general as TIEgen


def classifyTRACE(TRACES, pth=None):
    """Classify traces based on TIE parameters.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects to be classified.
    pth : list of float or None, optional
        Planarity thresholds for classification. If None, default thresholds
        [3, 9, 18] will be used. Must be of size 3.

    Returns
    -------
    list of trace objects (TIE_classes.trace_OBJ)
        The input list of trace objects with added classification parameters.
    """
    if pth is None:
        pth = [3, 9, 18]

    p1 = pth[0]
    p2 = pth[1]
    p3 = pth[2]

    for T in TRACES:
        Seg = T.Segment
        for S in Seg:
            aN = np.array(
                [
                    S.Chords[ci].alpha[0]
                    for ci in range(np.size(S.Chords))
                    if np.isnan(S.Chords[ci].alpha[0]) == False
                ]
            ).squeeze()
            aR = np.array(
                [
                    S.Chords[ci].alpha[1]
                    for ci in range(np.size(S.Chords))
                    if np.isnan(S.Chords[ci].alpha[1]) == False
                ]
            ).squeeze()
            bN = np.array(
                [
                    S.Chordplanes[cpi].beta[0]
                    for cpi in range(np.size(S.Chordplanes))
                    if np.isnan(S.Chordplanes[cpi].beta[0]) == False
                ]
            ).squeeze()
            bR = np.array(
                [
                    S.Chordplanes[cpi].beta[1]
                    for cpi in range(np.size(S.Chordplanes))
                    if np.isnan(S.Chordplanes[cpi].beta[1]) == False
                ]
            ).squeeze()

            alpS = np.sum(aN + aR)
            alpD = np.abs(np.sum(aN) - np.sum(aR))
            alp = alpS - alpD
            meana = alp / np.size(S.Chords)

            betS = np.sum(bN + bR)
            betD = np.abs(np.sum(bN) - np.sum(bR))
            bet = betS - betD
            meanb = bet / np.size(S.Chordplanes)
            if meanb == 0:
                meanb = 0.1

            meanr = meana / meanb
            fcurb1b = 180 / (meana - p1) + p1
            fcurb2b = 180 / (meana - p2) + p2
            fcurb3b = 180 / (meana - p3) + p3

            if meana < p1:
                g = 1
            elif meanb <= fcurb1b:
                g = 1
            elif meana < p2:
                g = 2
            elif meanb <= fcurb2b:
                g = 2
            elif meana < p3:
                g = 3
            elif meanb <= fcurb3b:
                g = 3
            elif meanb > fcurb3b:
                g = 4

            if meanr < 1:
                S.classID = -g
            else:
                S.classID = g

            S.sigheight = [meana, meanb]
        T.Segment = Seg
    return TRACES


def changeTHseg(TRACES, X, Y, Z, n, amp=-10, newTH=None):
    """
    Change Threshold For Segmentation.

    Changes the threshold that defines the convexity size to segment the traces
    for a specific trace n and performs the TIE on new segmentation.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    X, Y, Z : array_like
        Coordinate vectors.
    n : int
        Trace ID of the trace that is going to be analyzed with changed threshold.
    amp : float, optional
        Amplitude expressed in % by which the defined threshold should be
        lowered (negative) or increased (positive). Default is -10.
    newTH : list or None, optional
        Newly defined convexity size threshold [peak prominence, peak width].
        If newTH is defined, the amplitude is ignored. Default is None.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with adapted segmentation for the specific trace object.
    """
    if newTH is None:
        newTH = []  # Initialize the list here if it's None

    formthresh = TRACES[n - 1].convexityTH
    amp = (100 + amp) / 100
    if np.size(newTH) > 0:
        pthresh = newTH
    else:
        pthresh = [formthresh[0] * amp, formthresh[1] / amp]

    TRACES = clearTIE(TRACES, [n])
    TRACESn = tie(
        [TRACES[n - 1]], X, Y, Z, pth=[3, 9, 18], seg=True, peakthresh=pthresh
    )
    TRACES[n - 1] = TRACESn[0]
    TRACES[n - 1].convexityTH = pthresh

    return TRACES


def clearTIE(TRACES, n=None):
    """
    Clear TIE information from trace objects.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    n : list or None, optional
        List of trace IDs (integers) that designate the traces to be cleared.
        If n is None (default), all traces will be cleared.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with suppressed TIE information.
    """
    if n is None:
        n = []  # Initialize the list here if it's None

    if np.size(n) == 0:
        n = np.arange(0, int(np.size(TRACES))).astype(int)
    else:
        n = [ni - 1 for ni in n]

    for T in [TRACES[ni] for ni in n]:
        ltr = np.size(T.index)
        i_ind = np.arange(0, ltr).astype(int)
        seg = TIEclass.segment_OBJ(1, [], i_ind, i_ind[::-1], [], [], [], [])
        T.Segment = [seg]
        T.orientbar = []
        T.convexityTH = float("nan")

    return TRACES


def extractAlpha(TRACES):
    """
    Extracts alpha signal from traces.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with added alpha values for each trace object (respectively
        each chord object).
    """

    for T in TRACES:
        Segment = T.Segment
        for S in Segment:
            Chords = S.Chords
            n2 = np.size(Chords)
            vIniN = Chords[0].vector[0]  # initial vect for normal signal
            vIniR = Chords[0].vector[1]  # initial vect for reverse signal

            for i in range(n2):
                vAnaN = Chords[i].vector[0]  # analized vect for normal signal
                vAnaR = Chords[i].vector[1]  # analized vect for reverse signal
                angleN = TIEgen.angleBtwVec(vAnaN, vIniN)
                angleR = TIEgen.angleBtwVec(vAnaR, vIniR)
                Chords[i].alpha = [angleN, angleR]
            S.Chords = Chords
        T.Segment = Segment

    return TRACES


def extractBeta(TRACES):
    """
    Extracts beta signal from traces.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with added beta values for each trace object (respectively
        each chordPlane object).
    """

    for T in TRACES:
        Segment = T.Segment
        for S in Segment:
            ChdPlane = S.Chordplanes

            refpN = ChdPlane[0].normal[0]
            refpN = refpN / np.linalg.norm(refpN)
            refpR = ChdPlane[0].normal[1]
            refpR = refpR / np.linalg.norm(refpR)

            betN = [TIEgen.angleBtwVec(CP.normal[0], refpN) for CP in ChdPlane]
            betR = [TIEgen.angleBtwVec(CP.normal[1], refpR) for CP in ChdPlane]

            for j in range(1, np.size(betN)):
                if np.isnan(betN[j]) == False:
                    if betN[j] - betN[j - 1] > 90:
                        betN[j] = np.abs(betN[j] - 180)
                        betN[j] = betN[j]
            for j in range(1, np.size(betR)):
                if np.isnan(betR[j]) == False:
                    if betR[j] - betR[j - 1] > 90:
                        betR[j] = np.abs(betR[j] - 180)
                        betR[j] = betR[j]

            for i in range(np.size(betN)):
                ChdPlane[i].beta = [betN[i], betR[i]]

            S.Chordplanes = ChdPlane
        T.Segment = Segment
    return TRACES


def extractChdPlanes(TRACES, mX, mY, mZ):
    """
    Extracts chord planes from traces.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    mX, mY, mZ : array-like
        Matrices of coordinates of the analyzed extent.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with added list of chord plane objects (TIE_classes.chordPlane_OBJ)
        to each trace object (respectively each segment object).
    """

    mXfl = mX.flatten()
    mYfl = mY.flatten()
    mZfl = mZ.flatten()

    for T in TRACES:
        index = (T.index).astype(int)
        Segment = T.Segment

        tracel = np.zeros(np.size(index) - 1)
        for i in range(np.size(tracel)):
            PP1 = [mXfl[index[i]], mYfl[index[i]], mZfl[index[i]]]
            PP2 = [mXfl[index[i + 1]], mYfl[index[i + 1]], mZfl[index[i + 1]]]

            tracel[i] = TIEgen.distance(PP1, PP2)
        tr_length_sum = np.sum(tracel)

        for S in Segment:
            Chords = S.Chords
            N = np.size(Chords)

            if N % 2 == 0:
                step = int(N / 2)
                steppoint = step
            else:
                step = int((N - 1) / 2)
                steppoint = step + 1

            indexsegN = T.index[S.ind_normal].astype(int)  # Normal signal
            indexsegR = T.index[S.ind_reverse].astype(int)  # Reverse signal

            ChdPlane = []
            for i in range(steppoint):
                v1N = Chords[i].vector[0]
                v2N = Chords[i + step].vector[0]
                v1R = Chords[i].vector[1]
                v2R = Chords[i + step].vector[1]

                normalN = np.cross(v1N, v2N)
                azimN, dipN = TIEgen.normal2angle(normalN)
                normalR = np.cross(v1R, v2R)
                azimR, dipR = TIEgen.normal2angle(normalR)

                poleplN = 90 - dipN
                poletrN = azimN - 180
                if poletrN < 0:
                    poletrN = 360 + poletrN

                poleplR = 90 - dipR
                poletrR = azimR - 180
                if poletrR < 0:
                    poletrR = 360 + poletrR

                P1N = np.array(
                    [mXfl[indexsegN[i]], mYfl[indexsegN[i]], mZfl[indexsegN[i]]]
                )
                P2N = np.array(
                    [
                        mXfl[indexsegN[i + step]],
                        mYfl[indexsegN[i + step]],
                        mZfl[indexsegN[i + step]],
                    ]
                )

                P1P2N = P2N - P1N
                distN = np.abs(np.dot(np.cross(v1N, v2N), P1P2N)) / np.linalg.norm(
                    np.cross(v1N, v2N)
                )
                distratioN = distN / tr_length_sum

                P1R = np.array(
                    [mXfl[indexsegR[i]], mYfl[indexsegR[i]], mZfl[indexsegR[i]]]
                )
                P2R = np.array(
                    [
                        mXfl[indexsegR[i + step]],
                        mYfl[indexsegR[i + step]],
                        mZfl[indexsegR[i + step]],
                    ]
                )

                P1P2R = P2R - P1R
                distR = np.abs(np.dot(np.cross(v1R, v2R), P1P2R)) / np.linalg.norm(
                    np.cross(v1R, v2R)
                )
                distratioR = distR / tr_length_sum

                ChdPlane.append(
                    TIEclass.chordPlane_OBJ(
                        i + 1,
                        [normalN, normalR],
                        [azimN, dipN, azimR, dipR],
                        [poletrN, poleplN, poletrR, poleplR],
                        [],
                        [distN, distR],
                        [distratioN, distratioR],
                    )
                )
            S.Chordplanes = ChdPlane
        T.Segment = Segment

    return TRACES


def extractChords(TRACES, mX, mY, mZ):
    """
    Extracts connecting chords of a trace set (1st step of TIE).

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    mX, mY, mZ : array-like
        Matrices of coordinates of the analyzed extent.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with added list of chords (TIE_classes.chords_OBJ)
        to each trace object (respectively each segment object).
    """

    mXfl = mX.flatten()
    mYfl = mY.flatten()
    mZfl = mZ.flatten()

    for T in TRACES:
        indtr = (T.index).astype(int)
        Segment = T.Segment
        for S in Segment:
            l = np.size(S.ind_normal)
            if l % 2 > 0:
                DELTA = int((l - 1) / 2)
                lstr = DELTA + 1
            else:
                DELTA = int(l / 2)
                lstr = DELTA

            if lstr % 2 > 0:
                DD = int((lstr - 1) / 2)
            else:
                DD = int(lstr / 2)

            S.delta = [DELTA, DD]

            # normal trace
            chordsN = np.zeros([lstr, 3])
            trendN = np.zeros([lstr, 1])
            plungeN = np.zeros([lstr, 1])

            i = np.arange(0, lstr)
            pti = S.ind_normal[i]
            ptf = S.ind_normal[i + DELTA]
            chordsN[i, 0] = mXfl[indtr[pti]] - mXfl[indtr[ptf]]
            chordsN[i, 1] = mYfl[indtr[pti]] - mYfl[indtr[ptf]]
            chordsN[i, 2] = mZfl[indtr[pti]] - mZfl[indtr[ptf]]

            # Reverse trace
            chordsR = np.zeros([lstr, 3])
            trendR = np.zeros([lstr, 1])
            plungeR = np.zeros([lstr, 1])

            i = np.arange(0, lstr)
            pti = S.ind_reverse[i]
            ptf = S.ind_reverse[i + DELTA]
            chordsR[i, 0] = mXfl[indtr[pti]] - mXfl[indtr[ptf]]
            chordsR[i, 1] = mYfl[indtr[pti]] - mYfl[indtr[ptf]]
            chordsR[i, 2] = mZfl[indtr[pti]] - mZfl[indtr[ptf]]

            Chords = []
            for chd in range(lstr):
                trendN[chd], plungeN[chd] = TIEgen.vect2angle(chordsN[chd, :])
                trendR[chd], plungeR[chd] = TIEgen.vect2angle(chordsR[chd, :])
                Chords.append(
                    TIEclass.chord_OBJ(
                        chd + 1,
                        [chordsN[chd, :], chordsR[chd, :]],
                        [trendN[chd], trendR[chd]],
                        [plungeN[chd], plungeR[chd]],
                        [],
                    )
                )
            S.Chords = Chords
        T.Segment = Segment

    return TRACES


def extractOrientBars(TRACES):
    """
    Extraction of orientation bars (TIE).

    Extracts orientation bars from chord planes. Chord plane values are
    distributed along the trace.

    Parameters
    ----------
    TRACES : list of trace_OBJ
        List of trace objects containing basic TRACE information
        (any TRACE set, could also be FAULTS).

    Returns
    -------
    TRACES : list of trace_OBJ
        List with added orientation bar values to each trace object.
    """

    for T in TRACES:
        Segment = T.Segment
        or_new = np.zeros([np.size(T.index), 3])

        for S in Segment:
            bet = [S.Chordplanes[bi].beta[0] for bi in range(np.size(S.Chordplanes))]
            betI = [S.Chordplanes[bi].beta[1] for bi in range(np.size(S.Chordplanes))]

            if np.sum(bet) > np.sum(betI):
                orp = [
                    S.Chordplanes[bi].plane_orient[2:]
                    for bi in range(np.size(S.Chordplanes))
                ]
                si = S.ind_reverse
            else:
                orp = [
                    S.Chordplanes[bi].plane_orient[:2]
                    for bi in range(np.size(S.Chordplanes))
                ]
                si = S.ind_normal

            d = S.delta[0]
            dd = S.delta[1]

            for k in range(len(orp)):
                v1 = np.array(TIEgen.angle2vect(orp[k][0], orp[k][1])).squeeze()
                anchor_a = d + k + 1
                anchor_b = dd + d + k + 1
                vanch_a = slice(k, anchor_a)
                vanch_b = slice(dd + k, anchor_b)

                for a in si[vanch_a]:
                    or_new[a, :] = or_new[a, :] + v1
                for b in si[vanch_b]:
                    or_new[b, :] = or_new[b, :] + v1

        or_new = [m / np.linalg.norm(m) for m in or_new]
        for m in range(len(or_new)):
            if or_new[m][2] > 0:
                or_new[m] = or_new[m] * -1
        T.orientbar = or_new

    return TRACES


def mergeSeg(TRACES, n, s, X, Y, Z):
    """
    Merges two or more segments of a certain trace (trace_OBJ) into one single
    segment and reclassifies it according to TIE.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    n : int
        ID of specific trace_OBJ (only one trace_OBJ at a time).
    s : list
        List of segments that will be merged (e.g., [2, 3]).
    X, Y, Z : array-like
        Coordinate vectors.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with merged segments in the specific trace_OBJ.
    """

    s = [i - 1 for i in s]
    s.sort()
    for i in range(len(s) - 1):
        if abs(s[i] - s[i + 1]) > 1:
            TIEgen.Mbox("ERROR", "trace segments are not adjacent!!", 1)

    si = []
    # FIXME: list index out of range exception
    seglist = [TRACES[n - 1].Segment[ii] for ii in s]
    for k in seglist:
        si.extend(k.ind_normal.tolist())
    si = np.unique(si)

    SEG = TRACES[n - 1].Segment

    SEG[s[0]].ind_normal = si
    SEG[s[0]].ind_reverse = si[::-1]
    for i in range(1, len(s)):
        SEG.pop(int(s[i]))

    for sii in range(len(SEG)):
        SEG[sii].id = sii + 1

    TRACES[n - 1].Segment = SEG
    trace = tie([TRACES[n - 1]], X, Y, Z)
    TRACES[n - 1] = trace[0]
    TRACES[n - 1].convexityTH.append("merged " + str(s[0]) + " to " + str(s[-1]))

    return TRACES


def mergeTraces(TRACES, tn, X, Y, Z, seg=False):
    """
    Merges two or more traces into one single trace object
    and reclassifies it according to TIE.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    tn : list
        List of trace IDs that will be merged (e.g., [2, 3]).
    X, Y, Z : array-like
        Coordinate vectors.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with merged traces (new IDs).
    """

    tni = [ti - 1 for ti in tn]
    ni = []
    trlist = [TRACES[ii] for ii in tni]
    for k in trlist:
        ni.extend(k.index.tolist())
    ni = np.unique(ni).astype(int)
    ind_sort = TIEgen.sortLine(ni, np.shape(Z))

    TRACES[tni[0]].index = ind_sort.astype(int)
    TRACES = clearTIE(TRACES, [tn[0]])
    trace = tie([TRACES[tni[0]]], X, Y, Z, seg=seg)
    TRACES[tni[0]] = trace[0]

    for tri in range(len(TRACES)):
        if TRACES[tri].id in tn[1:]:
            TRACES[tri] = 1
    while 1 in TRACES:
        TRACES.remove(1)

    for tri in range(len(TRACES)):
        TRACES[tri].id = tri + 1

    return TRACES


def segSeg(TRACES, n, s, X, Y, Z):
    """
    Analyses a certain trace segment for potential segmentation at a
    more detailed scale. Important for long irregular traces that were not
    segmented during the first segmentation process.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    n : int
        ID of the trace object that contains the segment.
    s : int
        ID of the segment that will be re-analyzed for segmentation.
    X, Y, Z : array-like
        Coordinate vectors.

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with potentially added segments.
    """

    n = n - 1
    s = s - 1
    si = TRACES[n].Segment[s].ind_normal
    ni = TRACES[n].index[si].astype(int)
    TRACEX = copy.deepcopy(TRACES[n])
    TRACEX.index = copy.copy(ni)
    TRACEX = clearTIE([TRACEX])
    TRACEX = tie(TRACEX, X, Y, Z, seg=True)

    ls = len(TRACES[n].Segment)
    lsx = len(TRACEX[0].Segment)
    new_ls = ls + lsx - 1
    rest_i = np.arange(s + lsx, new_ls)
    new_i = np.arange(s, new_ls - np.size(rest_i))

    Seg = TRACES[n].Segment
    SegX = copy.deepcopy(TRACEX[0].Segment)

    Seg.pop(s)
    for k in range(np.size(new_i)):
        si_old = SegX[k].ind_normal
        si_new = si[SegX[k].ind_normal]
        Seg.insert(new_i[k], copy.deepcopy(SegX[k]))
        Seg[new_i[k]].ind_normal = si_new
        Seg[new_i[k]].ind_reverse = si_new[::-1]
        Seg[new_i[k]].id = new_i[k] + 1
        ind_slice = slice(si_new[0], si_new[-1] + 1, 1)
        ind_sliceX = slice(si_old[0], si_old[-1] + 1, 1)
        TRACES[n].orientbar[ind_slice] = TRACEX[0].orientbar[ind_sliceX]
    for k in rest_i:
        Seg[k].id = k + 1

    return TRACES


def segmentTRACE(TRACES, mX, mY, mZ, peakthresh=[100, 15]):
    """
    Segments the traces at their inflection points defined based on
    a peak threshold.

    Parameters
    ----------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List of trace objects.
    mX, mY, mZ : array-like
        Matrices of coordinates of the analyzed extent.
    peakthresh : list of float, optional
        Peak thresholds for convexity [peak prominence, peak width].
        Default is [100, 15].

    Returns
    -------
    TRACES : list of trace objects (TIE_classes.trace_OBJ)
        List with added segmentation where needed: list of
        segment objects (TIE_classes.segment_OBJ). Where no segmentation occurred,
        only one segment for one trace exists.
    """

    import scipy

    mXfl = mX.flatten()
    mYfl = mY.flatten()
    mZfl = mZ.flatten()

    for T in TRACES:
        ind = (T.index).astype(int)
        n = np.size(ind)
        d = 2  # steps between connecting points to create a vector v
        l = n - d  # final vector length of connected paths

        v = np.zeros([l, 3])
        angle = np.zeros([l, 1])

        if n > 5:
            k = np.arange(0, l)
            v[k, 0] = mXfl[ind[k + d]] - mXfl[ind[k]]
            v[k, 1] = mYfl[ind[k + d]] - mYfl[ind[k]]
            v[k, 2] = mZfl[ind[k + d]] - mZfl[ind[k]]

            angle = np.array([TIEgen.angleBtwVec(v[m, :], v[0, :]) for m in range(l)])
            angle = angle * np.pi / 180

            anmean1 = angle.copy()
            anmean2 = angle.copy()
            m = slice(1, l)
            p = slice(0, l - 1)
            anmean1[m] = (angle[m] + angle[p]) / 2
            anmean2[p] = (angle[m] + angle[p]) / 2

            # signal smoothing
            if n % 4 > 0:
                smo = int((n - (n % 4)) / 2)
            else:
                smo = int(n / 2)
            P = 1
            while P < smo:
                anmean1[m] = (anmean1[m] + anmean1[p]) / 2
                anmean2[p] = (anmean2[m] + anmean2[p]) / 2
                P = P + 1

            anmean = (
                np.cos(
                    np.concatenate(
                        (anmean1[int(smo / 2) :], anmean2[-smo : int(-smo / 2)])
                    )
                )
                * n
            )
            prom = peakthresh[0]  # minimal peak prominence
            width = n / peakthresh[1]  # minimal peak width

            # find signal peaks (positive and negative)
            ppi = scipy.signal.find_peaks(anmean, prominence=prom, width=width)
            npi = scipy.signal.find_peaks(
                -anmean + np.max(anmean), prominence=prom, width=width
            )
            pind = np.sort(np.concatenate(([0], ppi[0], npi[0], [l - 1])))
            pval = anmean[pind]

            # with each positive peak must follow a negative one -> peaks within peaks
            if np.size(pval) > 2:
                ploc = scipy.signal.find_peaks(pval)[0]
                nloc = scipy.signal.find_peaks(-pval)[0]
            else:
                ploc = []
                nloc = []

            seg = np.sort(np.concatenate((pind[ploc], pind[nloc])))

            # peak width also at the beginning and the end of the signal
            if np.size(seg) > 0:
                if seg[0] < 50:
                    if np.size(seg) > 1:
                        seg = seg[1:]
                    else:
                        seg = []

            if np.size(seg) > 0:
                if l - seg[-1] < 50:
                    if np.size(seg) > 1:
                        seg = seg[:-1]
                    else:
                        seg = []

            if np.size(seg) > 0:
                ind_normal = np.arange(0, seg[0] + d)
                T.Segment[0].ind_normal = ind_normal
                T.Segment[0].ind_reverse = ind_normal[::-1]

                if np.size(seg) > 1:
                    for s in range(1, np.size(seg)):
                        New_S = TIEclass.segment_OBJ(s + 1, [], [], [], [], [], [], [])
                        ind_normal = np.arange(seg[s - 1], seg[s] + d)
                        New_S.ind_normal = ind_normal
                        New_S.ind_reverse = ind_normal[::-1]
                        T.Segment.append(New_S)

                Last_S = TIEclass.segment_OBJ(
                    np.size(seg) + 1, [], [], [], [], [], [], []
                )
                ind_normal = np.arange(seg[-1], n)
                Last_S.ind_normal = ind_normal
                Last_S.ind_reverse = ind_normal[::-1]
                T.Segment.append(Last_S)

        T.convexityTH = peakthresh
    return TRACES


def tie(
    TRACES: List[TIEclass.trace_OBJ],
    X: np.ndarray,
    Y: np.ndarray,
    Z: np.ndarray,
    pth: List[int] = [3, 9, 18],
    seg: bool = False,
    peakthresh: List[int] = [100, 15],
) -> List[TIEclass.trace_OBJ]:
    """
    Performs the entire TIE analysis, step by step.

    Parameters
    ----------
    TRACES : list of trace_OBJ (TIE_classes.trace_OBJ)
        List of trace objects for TIE analysis.
    X, Y : numpy.ndarray
        Vectors of X and Y coordinates of the analyzed extent.
    Z : numpy.ndarray
        Matrix of Z values of the analyzed extent (size [len(X), len(Y)]).
    pth : list of int, optional
        Planarity thresholds for TIE classification. Default is [3, 9, 18].
    seg : bool, optional
        Boolean indicating if the trace should first undergo a segmentation
        process before extracting the trace information. Default is False.
    peakthresh : list of int, optional
        Peak thresholds for convexity [peak prominence, peak width]. Default
        is [90, 15], and it is not used if seg=False.

    Returns
    -------
    TRACES : list of trace_OBJ
        The same list with added TIE information.
    """

    [mX, mY] = np.meshgrid(X, Y)
    mX = np.fliplr(mX)
    mZ = np.fliplr(Z)
    if seg == True:
        TRACES = segmentTRACE(TRACES, mX, mY, mZ, peakthresh)

    TRACES = extractChords(TRACES, mX, mY, mZ)
    TRACES = extractAlpha(TRACES)
    TRACES = extractChdPlanes(TRACES, mX, mY, mZ)
    TRACES = extractBeta(TRACES)
    TRACES = classifyTRACE(TRACES, pth)
    TRACES = extractOrientBars(TRACES)

    return TRACES
